import cPickle
import time
import pprint
import re
from twisted.web.client import _parse
from uuid import uuid5, NAMESPACE_DNS
from twisted.internet.defer import Deferred, DeferredList, maybeDeferred
from twisted.internet import task
from twisted.internet import reactor
from twisted.web import server
from .base import BaseServer, LOGGER
from ..aws import sdb_now, sdb_now_add
from ..resources import ExecutionResource
from ..networkaddress import getNetworkAddress

PRETTYPRINTER = pprint.PrettyPrinter(indent=4)

class ExecutionServer(BaseServer):
    
    peers = {}
    peer_uuids = []
    reportjobspeedloop = None
    jobsloop = None
    queryloop = None
    coordinateloop = None  
    uuid_limits = {'start':None, 'end':None}
    public_ip = None
    local_ip = None
    network_information = {}
    queued_jobs = {}
    job_queue = []
    job_count = 0
    query_start_time = None
    simultaneous_jobs = 30
    querying_for_jobs = False
    reservation_update_queue = []
    current_sql = ""
    last_job_query_count = 0
    def __init__(self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_sdb_reservation_domain, 
            aws_s3_http_cache_bucket=None, 
            aws_s3_storage_bucket=None,
            aws_sdb_coordination_domain=None,
            aws_s3_reservation_cache_bucket=None,
            max_simultaneous_requests=50,
            max_requests_per_host_per_second=0,
            max_simultaneous_requests_per_host=0,
            port=5001, 
            log_file='executionserver.log',
            log_directory=None,
            log_level="debug",
            name=None,
            time_offset=None,
            peer_check_interval=60,
            reservation_check_interval=60,
            hammer_prevention=False):
        if name == None:
            name = "AWSpider Execution Server UUID: %s" % self.uuid
        self.network_information["port"] = port
        self.hammer_prevention = hammer_prevention
        self.peer_check_interval = int(peer_check_interval)
        self.reservation_check_interval = int(reservation_check_interval)
        resource = ExecutionResource(self)
        self.site_port = reactor.listenTCP(port, server.Site(resource))
        BaseServer.__init__(
            self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_s3_reservation_cache_bucket=aws_s3_reservation_cache_bucket,
            aws_s3_http_cache_bucket=aws_s3_http_cache_bucket, 
            aws_sdb_reservation_domain=aws_sdb_reservation_domain, 
            aws_s3_storage_bucket=aws_s3_storage_bucket,
            aws_sdb_coordination_domain=aws_sdb_coordination_domain,
            max_simultaneous_requests=max_simultaneous_requests,
            max_requests_per_host_per_second=max_requests_per_host_per_second,
            max_simultaneous_requests_per_host=max_simultaneous_requests_per_host,
            log_file=log_file,
            log_directory=log_directory,
            log_level=log_level,
            name=name,
            time_offset=time_offset,
            port=port)

    def start(self):
        reactor.callWhenRunning(self._start)
        return self.start_deferred

    def _start(self):
        deferreds = []
        deferreds.append(self.getNetworkAddress())
        if self.time_offset is None:
            deferreds.append(self.getTimeOffset())
        d = DeferredList(deferreds, consumeErrors=True)
        d.addCallback(self._startCallback)

    def _startCallback(self, data):
        for row in data:
            if row[0] == False:
                d = self.shutdown()
                d.addCallback(self._startHandleError, row[1])
                return d
        d = BaseServer.start(self)   
        d.addCallback(self._startCallback2)

    def _startCallback2(self, data):
        if self.shutdown_trigger_id is not None:
            self.reportjobspeedloop = task.LoopingCall(self.reportJobSpeed)
            self.reportjobspeedloop.start(60)
            self.jobsloop = task.LoopingCall(self.executeJobs)
            self.jobsloop.start(1)
            if self.aws_sdb_coordination_domain is not None:
                self.peerCheckRequest()  
                d = maybeDeferred(self.coordinate)
                d.addCallback(self._startCallback3)
            else:
                self.queryloop = task.LoopingCall(self.query)
                self.queryloop.start(self.reservation_check_interval)
    
    def _startCallback3(self, data):
        self.coordinateloop = task.LoopingCall(self.coordinate)
        self.coordinateloop.start(self.peer_check_interval)
        self.queryloop = task.LoopingCall(self.query)
        self.queryloop.start(self.reservation_check_interval)
    
    def shutdown(self):
        LOGGER.critical("Shutting down.")
        self.job_queue = []
        self.queued_jobs = {}
        deferreds = []
        LOGGER.debug("%s stopping on main HTTP interface." % self.name)
        d = self.site_port.stopListening()
        if isinstance(d, Deferred):
            deferreds.append(d)
        if self.reportjobspeedloop is not None:
            LOGGER.debug("Stopping report job speed loop.")
            d = self.reportjobspeedloop.stop()
            if isinstance(d, Deferred):
                deferreds.append(d)            
        if self.jobsloop is not None:
            LOGGER.debug("Stopping jobs loop.")
            d = self.jobsloop.stop()
            if isinstance(d, Deferred):
                deferreds.append(d)           
        if self.queryloop is not None:
            LOGGER.debug("Stopping query loop.")
            d = self.queryloop.stop()
            if isinstance(d, Deferred):
                deferreds.append(d)
        if self.coordinateloop is not None:
            LOGGER.debug("Stopping coordinating loop.")
            d = self.coordinateloop.stop()
            if isinstance(d, Deferred):
                deferreds.append(d)
            LOGGER.debug("Removing data from SDB coordination domain.")
            d = self.sdb.delete(self.aws_sdb_coordination_domain, self.uuid)
            d.addCallback(self.peerCheckRequest)
            deferreds.append(d)
        if len(deferreds) > 0:
            d = DeferredList(deferreds)
            d.addCallback(self._shutdownCallback)
            return d
        else:
            return self._shutdownCallback(None)
    
    def _shutdownCallback(self, data):
        return BaseServer.shutdown(self)

    def peerCheckRequest(self, data=None):
        LOGGER.debug("Signaling peers.")
        deferreds = []
        for uuid in self.peers:
            if uuid != self.uuid and self.peers[uuid]["active"]:
                LOGGER.debug("Signaling %s to check peers." % self.peers[uuid]["uri"])
                d = self.rq.getPage(
                    self.peers[uuid]["uri"] + "/coordinate", 
                    prioritize=True)
                d.addCallback(
                    self._peerCheckRequestCallback, 
                    self.peers[uuid]["uri"])
                d.addErrback(
                    self._peerCheckRequestErrback, 
                    self.peers[uuid]["uri"])
                deferreds.append(d)
        if len(deferreds) > 0:
            LOGGER.debug("Combinining shutdown signal deferreds.")
            return DeferredList(deferreds, consumeErrors=True)
        return True
    
    def _peerCheckRequestErrback(self, error, uri):
        LOGGER.debug("Could not get %s/coordinate: %s" % (uri, str(error)))
        
    def _peerCheckRequestCallback(self, data, uri):
        LOGGER.debug("Got %s/coordinate." % uri)

    def getNetworkAddress(self):
        d = getNetworkAddress()
        d.addCallback(self._getNetworkAddressCallback)
        d.addErrback(self._getNetworkAddressErrback)
        return d

    def _getNetworkAddressCallback(self, data):   
        if "public_ip" in data:
            self.public_ip = data["public_ip"]
            self.network_information["public_ip"] = self.public_ip
        if "local_ip" in data:
            self.local_ip = data["local_ip"]  
            self.network_information["local_ip"] = self.local_ip

    def _getNetworkAddressErrback(self, error):
        message = "Could not get network address." 
        LOGGER.error(message)
        raise Exception(message)

    def coordinate(self):
        server_data = self.getServerData()
        attributes = {
            "created":sdb_now(offset=self.time_offset),
            "load_avg":server_data["load_avg"],
            "running_time":server_data["running_time"],
            "cost":server_data["cost"],
            "active_requests":server_data["active_requests"],
            "pending_requests":server_data["pending_requests"],
            "current_timestamp":server_data["current_timestamp"],
            "job_queue":len(self.job_queue),
            "active_jobs":len(self.active_jobs),
            "queued_jobs":len(self.queued_jobs),
            "current_sql":self.current_sql.replace("\n", ""),
            "last_job_query_count":self.last_job_query_count}
        if self.uuid_limits["start"] is None and self.uuid_limits["end"] is not None:
            attributes["range"] = "Start - %s" % self.uuid_limits["end"]
        elif self.uuid_limits["start"] is not None and self.uuid_limits["end"] is None:
            attributes["range"] = "%s - End" % self.uuid_limits["start"]
        elif self.uuid_limits["start"] is None and self.uuid_limits["end"] is None:
            attributes["range"] = "Full range"
        else:
            attributes["range"] = "%s - %s" % (self.uuid_limits["start"], self.uuid_limits["end"])
        attributes.update(self.network_information)
        d = self.sdb.putAttributes(
            self.aws_sdb_coordination_domain, 
            self.uuid, 
            attributes, 
            replace=attributes.keys())
        d.addCallback(self._coordinateCallback)
        d.addErrback(self._coordinateErrback)
        return d
        
    def _coordinateCallback(self, data):
        sql = "SELECT public_ip, local_ip, port FROM `%s` WHERE created > '%s'" % (
            self.aws_sdb_coordination_domain, 
            sdb_now_add(self.peer_check_interval * -2, 
            offset=self.time_offset))
        LOGGER.debug("Querying SimpleDB, \"%s\"" % sql)
        d = self.sdb.select(sql)
        d.addCallback(self._coordinateCallback2)
        d.addErrback(self._coordinateErrback)
        return d
        
    def _coordinateCallback2(self, discovered):
        existing_peers = set(self.peers.keys())
        discovered_peers = set(discovered.keys())
        new_peers = discovered_peers - existing_peers
        old_peers = existing_peers - discovered_peers
        for uuid in old_peers:
            LOGGER.debug("Removing peer %s" % uuid)
            if uuid in self.peers:
                del self.peers[uuid]
        deferreds = []
        for uuid in new_peers:
            if uuid == self.uuid:
                self.peers[uuid] = {
                    "uri":"http://127.0.0.1:%s" % self.port,
                    "local_ip":"127.0.0.1",
                    "port":self.port,
                    "active":True
                }
            else:
                deferreds.append(self.verifyPeer(uuid, discovered[uuid]))
        if len(new_peers) > 0:
            if len(deferreds) > 0:
                d = DeferredList(deferreds, consumeErrors=True)
                d.addCallback(self._coordinateCallback3)
                return d
            else:
                self._coordinateCallback3(None) #Just found ourself.
        elif len(old_peers) > 0:
            self._coordinateCallback3(None)
        else:
            pass # No old, no new.

    def _coordinateCallback3(self, data):
        LOGGER.debug("Re-organizing peers.")
        for uuid in self.peers:
            if "local_ip" in self.peers[uuid]:
                self.peers[uuid]["uri"] = "http://%s:%s" % (
                    self.peers[uuid]["local_ip"], 
                    self.peers[uuid]["port"])
                self.peers[uuid]["active"] = True
                self.rq.setHostMaxRequestsPerSecond(
                    self.peers[uuid]["local_ip"], 
                    0)
                self.rq.setHostMaxSimultaneousRequests(
                    self.peers[uuid]["local_ip"], 
                    0)
            elif "public_ip" in self.peers[uuid]:
                self.peers[uuid]["uri"] = "http://%s:%s" % (
                    self.peers[uuid]["public_ip"], 
                    self.peers[uuid]["port"])
                self.peers[uuid]["active"] = True
                self.rq.setHostMaxRequestsPerSecond(
                    self.peers[uuid]["public_ip"], 
                    0)
                self.rq.setHostMaxSimultaneousRequests(
                    self.peers[uuid]["public_ip"],
                    0)
            else:
                LOGGER.error("Peer %s has no local or public IP. This should not happen." % uuid)
        self.peer_uuids = self.peers.keys()
        self.peer_uuids.sort()
        LOGGER.debug("Peers updated to: %s" % self.peers)
        # Set UUID peer limits by splitting up lexicographical namespace using hex values.
        peer_count = len(self.peers)
        splits = [hex(4096/peer_count * x)[2:] for x in range(1, peer_count)]
        splits = zip([None] + splits, splits + [None])
        splits = [{"start":x[0], "end":x[1]} for x in splits]
        if self.uuid in self.peer_uuids:
            self.uuid_limits = splits[self.peer_uuids.index(self.uuid)]
        else:
            self.uuid_limits = {"start":None, "end":None}
        job_queue_length = len(self.job_queue)
        if self.uuid_limits["start"] is None and self.uuid_limits["end"] is not None:
            self.job_queue = filter(self.testJobByEnd, self.job_queue)
        elif self.uuid_limits["start"] is not None and self.uuid_limits["end"] is None:
            self.job_queue = filter(self.testJobByStart, self.job_queue)
        elif self.uuid_limits["start"] is not None and self.uuid_limits["end"] is not None:
            self.job_queue = filter(self.testJobByStartAndEnd, self.job_queue)
        self.queued_jobs = dict((x["uuid"], True) for x in self.job_queue)
        LOGGER.info("Abandoned %s jobs that were out of range." % (job_queue_length - len(self.job_queue)))
        LOGGER.debug("Updated UUID limits to: %s" % self.uuid_limits)
    
    def _coordinateErrback(self, error):
        LOGGER.error("Could not query SimpleDB for peers: %s" % str(error))

    def testJobByEnd(self, job):
        return job["uuid"] < self.uuid_limits["end"]

    def testJobByStart(self, job):
        return job["uuid"] > self.uuid_limits["start"]

    def testJobByStartAndEnd(self, job):
        return (job["uuid"] < self.uuid_limits["end"] and job["uuid"] > self.uuid_limits["start"])
        
    def verifyPeer(self, uuid, peer):
        LOGGER.debug("Verifying peer %s" % uuid)
        deferreds = []
        if "port" in peer:
            port = int(peer["port"][0])
        else:
            port = self.port
        if uuid not in self.peers:
            self.peers[uuid] = {}
        self.peers[uuid]["active"] = False
        self.peers[uuid]["port"] = port
        if "local_ip" in peer:
            local_ip = peer["local_ip"][0]
            local_url = "http://%s:%s/server" % (local_ip, port)
            d = self.rq.getPage(local_url, timeout=5, prioritize=True)
            d.addCallback(self._verifyPeerLocalIPCallback, uuid, local_ip, port)
            deferreds.append(d)
        if "public_ip" in peer:
            public_ip = peer["public_ip"][0]
            public_url = "http://%s:%s/server" % (public_ip, port)
            d = self.rq.getPage(public_url, timeout=5, prioritize=True)
            d.addCallback(self._verifyPeerPublicIPCallback, uuid, public_ip, port)
            deferreds.append(d)
        if len(deferreds) > 0:
            d = DeferredList(deferreds, consumeErrors=True)
            return d
        else:
            return None

    def _verifyPeerLocalIPCallback(self, data, uuid, local_ip, port):
        LOGGER.debug("Verified local IP for %s" % uuid)
        self.peers[uuid]["local_ip"] = local_ip

    def _verifyPeerPublicIPCallback(self, data, uuid, public_ip, port):
        LOGGER.debug("Verified public IP for %s" % uuid)
        self.peers[uuid]["public_ip"] = public_ip

    def getPage(self, *args, **kwargs):
        if not self.hammer_prevention or len(self.peer_uuids) == 0:
            return self.pg.getPage(*args, **kwargs)
        else:
            scheme, host, port, path = _parse(args[0])
            peer_key = int(uuid5(NAMESPACE_DNS, host).int % len(self.peer_uuids))
            peer_uuid = self.peer_uuids[peer_key]
            if peer_uuid == self.uuid or self.peers[peer_uuid]["active"] == False:
                return self.pg.getPage(*args, **kwargs)
            else:
                parameters = {}
                parameters["url"] = args[0]
                if "method" in kwargs:
                    parameters["method"] = kwargs["method"]   
                if "postdata" in kwargs: 
                    parameters["postdata"] = urllib.urlencode(kwargs["postdata"])
                if "headers" in kwargs: 
                    parameters["headers"] = urllib.urlencode(kwargs["headers"])
                if "cookies" in kwargs: 
                    parameters["cookies"] = urllib.urlencode(kwargs["cookies"])         
                if "agent" in kwargs:
                    parameters["agent"] = kwargs["agent"]
                if "timeout" in kwargs:
                    parameters["timeout"] = kwargs["timeout"]
                if "followRedirect" in kwargs:
                    parameters["followRedirect"] = kwargs["followRedirect"]
                if "url_hash" in kwargs: 
                    parameters["url_hash"] = kwargs["url_hash"]
                if "cache" in kwargs: 
                    parameters["cache"] = kwargs["cache"]
                if "prioritize" in kwargs: 
                    parameters["prioritize"] = kwargs["prioritize"]
                url = "%s/getpage?%s" % (
                    self.peers[peer_uuid]["uri"], 
                    urllib.urlencode(parameters))
                LOGGER.debug("Rerouting request for %s to %s" % (args[0], url))
                d = self.rq.getPage(url, prioritize=True)
                d.addErrback(self._getPageErrback, args, kwargs) 
                return d

    def _getPageErrback(self, error, args, kwargs):
        LOGGER.error(args[0] + ":" + str(error))
        return self.pg.getPage(*args, **kwargs)

    def queryByUUID(self, uuid):
        sql = "SELECT * FROM `%s` WHERE itemName() = '%s'" % (
            self.aws_sdb_reservation_domain, 
            uuid)
        LOGGER.debug("Querying SimpleDB, \"%s\"" % sql)
        d = self.sdb.select(sql)
        d.addCallback(self._queryCallback2)
        d.addErrback(self._queryErrback)        
        return d

    def query(self, data=None):
        if len(self.job_queue) > 1000:
            LOGGER.debug("Skipping query. %s jobs already active." % len(self.job_queue))
            return
        if self.querying_for_jobs:
            LOGGER.debug("Skipping query. Already querying for jobs.")
            return
        self.querying_for_jobs = True
        if self.uuid_limits["start"] is None and self.uuid_limits["end"] is not None:
            uuid_limit_clause = "AND itemName() < '%s'" % self.uuid_limits["end"]
        elif self.uuid_limits["start"] is not None and self.uuid_limits["end"] is None:
            uuid_limit_clause = "AND itemName() > '%s'" % self.uuid_limits["start"]
        elif self.uuid_limits["start"] is None and self.uuid_limits["end"] is None:
            uuid_limit_clause = ""
        else:
            uuid_limit_clause = "AND itemName() BETWEEN '%s' AND '%s'" % (
                self.uuid_limits["start"], 
                self.uuid_limits["end"])
        sql = """SELECT * 
                FROM `%s` 
                WHERE
                reservation_next_request < '%s' %s
                LIMIT 2500""" % (
                self.aws_sdb_reservation_domain, 
                sdb_now(offset=self.time_offset),
                uuid_limit_clause)
        sql = re.sub(r"\s\s*", " ", sql);
        self.current_sql = sql
        LOGGER.debug("Querying SimpleDB, \"%s\"" % sql)
        d = self.sdb.select(sql, max_results=5000)
        d.addCallback(self._queryCallback)
        d.addErrback(self._queryErrback)

    def _queryErrback(self, error):
        self.querying_for_jobs = False
        LOGGER.error("Unable to query SimpleDB.\n%s" % error)
        
    def _queryCallback(self, data):
        LOGGER.info("Fetched %s jobs." % len(data))
        self.querying_for_jobs = False
        # Iterate through the reservation data returned from SimpleDB
        self.last_job_query_count = len(data)
        for uuid in data:
            if uuid in self.active_jobs or uuid in self.queued_jobs:
                continue
            kwargs_raw = {}
            reserved_arguments = {}
            # Load attributes into dicts for use by the system or custom functions.
            for key in data[uuid]:
                if key in self.reserved_arguments:
                    reserved_arguments[key] = data[uuid][key][0]
                else:
                    kwargs_raw[key] = data[uuid][key][0]
            # Check for the presence of all required system attributes.
            if "reservation_function_name" not in reserved_arguments:
                LOGGER.error("Reservation %s does not have a function name." % uuid)
                self.deleteReservation(uuid)
                continue
            function_name = reserved_arguments["reservation_function_name"]
            if function_name not in self.functions:
                LOGGER.error("Unable to process function %s for UUID: %s" % (function_name, uuid))
                continue
            if "reservation_created" not in reserved_arguments:
                LOGGER.error("Reservation %s, %s does not have a created time." % (function_name, uuid))
                self.deleteReservation(uuid, function_name=function_name)
                continue
            if "reservation_next_request" not in reserved_arguments:
                LOGGER.error("Reservation %s, %s does not have a next request time." % (function_name, uuid))
                self.deleteReservation(uuid, function_name=function_name)
                continue                
            if "reservation_error" not in reserved_arguments:
                LOGGER.error("Reservation %s, %s does not have an error flag." % (function_name, uuid))
                self.deleteReservation(uuid, function_name=function_name)
                continue
            # Load custom function.
            if function_name in self.functions:
                exposed_function = self.functions[function_name]
            else:
                LOGGER.error("Could not find function %s." % function_name)
                continue
            # Check for required / optional arguments.
            kwargs = {}
            for key in kwargs_raw:
                if key in exposed_function["required_arguments"]:
                    kwargs[key] = kwargs_raw[key]
                if key in exposed_function["optional_arguments"]:
                    kwargs[key] = kwargs_raw[key]
            has_required_arguments = True
            for key in exposed_function["required_arguments"]:
                if key not in kwargs:
                    has_required_arguments = False
                    LOGGER.error("%s, %s does not have required argument %s." % (function_name, uuid, key))
            if not has_required_arguments:
                continue
            self.queued_jobs[uuid] = True
            job = {"exposed_function":exposed_function,
                "kwargs":kwargs,
                "function_name":function_name,
                "uuid":uuid}
            if "reservation_cache" in reserved_arguments:
                LOGGER.debug("Using reservation fast cache for %s, %s on on SimpleDB." % (function_name, uuid))
                job["reservation_cache"] = reserved_arguments["reservation_cache"]
            else:
                job["reservation_cache"] = None
            self.job_queue.append(job)
        self.job_count = 0
        self.query_start_time = time.time()
        self.executeJobs()            

    def reportJobSpeed(self):
        if self.query_start_time is not None and self.job_count > 0:
            seconds_per_job = (time.time() - self.query_start_time) / self.job_count
            LOGGER.info("Average execution time: %s, %s active." % (seconds_per_job, len(self.active_jobs)))
        else:
            LOGGER.info("No average speed to report yet.")
            
    def executeJobs(self, data=None):           
        while len(self.job_queue) > 0 and len(self.active_jobs) < self.simultaneous_jobs:
            job = self.job_queue.pop(0)
            exposed_function = job["exposed_function"]
            kwargs = job["kwargs"]
            function_name = job["function_name"]
            uuid = job["uuid"]
            del self.queued_jobs[uuid]
            LOGGER.debug("Calling %s with args %s" % (function_name, kwargs))
            d = self.callExposedFunction(
                exposed_function["function"], 
                kwargs, 
                function_name, 
                uuid=uuid,
                reservation_fast_cache=job["reservation_cache"])
            d.addCallback(self._jobCountCallback)
            d.addErrback(self._jobCountErrback)
            d.addCallback(self._setNextRequest, uuid, exposed_function["interval"], function_name)
        
    def _jobCountCallback(self, data=None):
        self.job_count += 1
        
    def _jobCountErrback(self, error):
        self.job_count += 1
    
    def _setNextRequest(self, data, uuid, exposed_function_interval, function_name):
        reservation_next_request_parameters = {
            "reservation_next_request":sdb_now_add(
                exposed_function_interval, 
                offset=self.time_offset)}
        if uuid in self.reservation_fast_caches:
            LOGGER.debug("Set reservation fast cache for %s, %s on on SimpleDB." % (function_name, uuid))
            reservation_next_request_parameters["reservation_cache"] = self.reservation_fast_caches[uuid]
            del self.reservation_fast_caches[uuid]
        self.reservation_update_queue.append((
            uuid,
            reservation_next_request_parameters))
        if len(self.reservation_update_queue) > 25:
            self._sendReservationUpdateQueue()
    
    def _sendReservationUpdateQueue(self, data=None):
        if len(self.reservation_update_queue) < 25 and (len(self.active_jobs) > 0 or len(self.job_queue) > 0):
            return
        if len(self.reservation_update_queue) == 0:
            return
        LOGGER.debug("Sending reservation queue. Current length is %s" % (
            len(self.reservation_update_queue)))
        reservation_updates = self.reservation_update_queue[0:25]
        replace=["reservation_next_request", "reservation_cache"]
        reservation_replaces = [(x[0], replace) for x in reservation_updates]
        reservation_updates = dict(reservation_updates)
        reservation_replaces = dict(reservation_replaces)
        self.reservation_update_queue = self.reservation_update_queue[25:]
        d = self.sdb.batchPutAttributes(
            self.aws_sdb_reservation_domain,
            reservation_updates,
            replace_by_item_name=reservation_replaces)
        d.addCallback(
            self._sendReservationUpdateQueueCallback, 
            reservation_updates.keys())
        d.addErrback(
            self._sendReservationUpdateQueueErrback, 
            reservation_updates.keys(),
            reservation_updates)
        if len(self.reservation_update_queue) > 0:
            reactor.callLater(0, self._sendReservationUpdateQueue)

    def _sendReservationUpdateQueueCallback(self, data, uuids):
        LOGGER.debug("Set next request for %s on on SimpleDB." % uuids)
    
    def _sendReservationUpdateQueueErrback(self, error, uuids, reservation_updates): 
        LOGGER.error("Unable to set next request for %s on SimpleDB. Adding back to update queue.\n%s" % (uuids, error.value))
        self.reservation_update_queue.extend(reservation_updates.items())



from twisted.web import server
from .base import BaseResource


class DataResource(BaseResource):
    
    def __init__(self, dataserver):
        self.dataserver = dataserver
        BaseResource.__init__(self)
    
    def render(self, request):
        request.setHeader('Content-type', 'text/javascript; charset=UTF-8')
        if "uuid" in request.args:
            get_deferreds = []
            for uuid in request.args["uuid"]:
                get_deferreds.append(self.dataserver.getData(uuid))
            d = DeferredList(get_deferreds, consumeErrors = True)
            d.addCallback(self._getCallback, request.args["uuid"])
            d.addCallback(self._successResponse)
            d.addErrback(self._errorResponse)
            d.addCallback(self._immediateResponse, request)
            return server.NOT_DONE_YET
        return self._errorResponse(Failure(exc_value=Exception("Parameter UUID is required."))) 

    def _getCallback(self, data, uuids):
        response = {}
        for i in range(0, len(uuids)):
            if data[i][0] == True:
                response[uuids[i]] = data[i][1]
            else:
                response[uuids[i]] = {"error":str(data[i][1].value)}
        return response




from uuid import UUID, uuid4
import time
import random
import logging
import logging.handlers
from heapq import heappush, heappop
from twisted.internet import reactor, task
from twisted.web import server
from twisted.enterprise import adbapi
from MySQLdb.cursors import DictCursor
from twisted.internet.defer import Deferred, inlineCallbacks, DeferredList
from twisted.internet import task
from twisted.internet.threads import deferToThread
from txamqp.content import Content
from .base import BaseServer, LOGGER
from ..resources2 import SchedulerResource
from ..amqp import amqp as AMQP
from ..resources import ExposedResource
from twisted.web.resource import Resource


class SchedulerServer(BaseServer):
    exposed_functions = []
    exposed_function_resources = {}
    name = "AWSpider Schedule Server UUID: %s" % str(uuid4())
    heap = []
    enqueueCallLater = None
    
    def __init__(self,
            mysql_username,
            mysql_password,
            mysql_host,
            mysql_database,
            amqp_host,
            amqp_username,
            amqp_password,
            amqp_vhost,
            amqp_queue,
            amqp_exchange,
            amqp_port=5672,
            mysql_port=3306,
            port=5004, 
            service_mapping=None,
            log_file='schedulerserver.log',
            log_directory=None,
            log_level="debug"):
        self.function_resource = Resource()
        # Create MySQL connection.
        self.mysql = adbapi.ConnectionPool(
            "MySQLdb", 
            db=mysql_database, 
            port=mysql_port, 
            user=mysql_username, 
            passwd=mysql_password, 
            host=mysql_host, 
            cp_reconnect=True, 
            cursorclass=DictCursor)
        # Resource Mappings
        self.service_mapping = service_mapping
        # AMQP connection parameters
        self.amqp_host = amqp_host
        self.amqp_vhost = amqp_vhost
        self.amqp_port = amqp_port
        self.amqp_username = amqp_username
        self.amqp_password = amqp_password
        self.amqp_queue = amqp_queue
        self.amqp_exchange = amqp_exchange
        # HTTP interface
        resource = SchedulerResource(self)
        self.function_resource = Resource()
        resource.putChild("function", self.function_resource)
        self.site_port = reactor.listenTCP(port, server.Site(resource))
        # Logging, etc
        BaseServer.__init__(
            self,
            log_file=log_file,
            log_directory=log_directory,
            log_level=log_level) 
        function_name = BaseServer.makeCallable(
            self, 
            self.remoteAddToHeap, 
            interval=0, 
            name=None, 
            expose=True)
        er = ExposedResource(self, function_name)
        function_name_parts = function_name.split("/")
        if len(function_name_parts) > 1:
            if function_name_parts[0] in self.exposed_function_resources:
                r = self.exposed_function_resources[function_name_parts[0]]
            else:
                r = Resource()
                self.exposed_function_resources[function_name_parts[0]] = r
            self.function_resource.putChild(function_name_parts[0], r)
            r.putChild(function_name_parts[1], er)
        else:
            self.function_resource.putChild(function_name_parts[0], er)
        LOGGER.info("Function %s is now available via the HTTP interface." % function_name)
    
    def start(self):
        reactor.callWhenRunning(self._start)
        return self.start_deferred
    
    @inlineCallbacks   
    def _start(self):
        # Load in names of functions supported by plugins
        self.function_names = self.functions.keys()
        LOGGER.info('Connecting to broker.')
        self.conn = yield AMQP.createClient(
            self.amqp_host, 
            self.amqp_vhost, 
            self.amqp_port)
        yield self.conn.authenticate(self.amqp_username, self.amqp_password)
        self.chan = yield self.conn.channel(1)
        yield self.chan.channel_open()
        # Create Queue
        yield self.chan.queue_declare(
            queue=self.amqp_queue, 
            durable=False, 
            exclusive=False, 
            auto_delete=False)
        # Create Exchange
        yield self.chan.exchange_declare(
            exchange=self.amqp_exchange, 
            type="fanout", 
            durable=False, 
            auto_delete=False)
        yield self.chan.queue_bind(
            queue=self.amqp_queue, 
            exchange=self.amqp_exchange)
        # Build heap from data in MySQL
        yield self._loadFromMySQL()
        
    def _loadFromMySQL(self, start=0):
        # Select the entire spider_service DB, 10k rows at at time.
        sql = "SELECT uuid, type FROM spider_service ORDER BY id LIMIT %s, 10000" % start
        LOGGER.debug(sql)
        d = self.mysql.runQuery(sql)
        d.addCallback(self._loadFromMySQLCallback, start)
        d.addErrback(self._loadFromMySQLErrback)
        return d
        
    def _loadFromMySQLCallback(self, data, start):
        # Add rows to heap. The second argument is interval, would be 
        # based on the plugin's interval setting, random for now.
        for row in data:
            self.addToHeap(row["uuid"], row["type"])
        # Load next chunk.
        #if len(data) >= 10000:
        #    return self._loadFromMySQL(start=start + 10000)
        # Done loading, start queuing
        self.enqueue()
        d = BaseServer.start(self)   
        return d
    
    def _loadFromMySQLErrback(self, error):
        return error
    
    @inlineCallbacks
    def shutdown(self):
        LOGGER.debug("Closting connection")
        try:
            self.enqueueCallLater.cancel()
        except:
            pass
        # Shut things down
        LOGGER.info('Closing broker connection')
        yield self.chan.channel_close()
        chan0 = yield self.conn.channel(0)
        yield chan0.connection_close()
    
    def enqueue(self):
        # Defer this to a thread so we don't block on the web interface.
        deferToThread(self._enqueue)
        
    def _enqueue(self):
        now = int(time.time())
        # Compare the heap min timestamp with now().
        # If it's time for the item to be queued, pop it, update the 
        # timestamp and add it back to the heap for the next go round.
        queue_items = []
        queue_items_a = queue_items.append
        LOGGER.debug("%s:%s" % (self.heap[0][0], now))
        while self.heap[0][0] < now:
            job = heappop(self.heap)
            queue_items_a(job[1][0])
            new_job = (now + job[1][1], job[1])
            heappush(self.heap, new_job)
        # add items to amqp
        if queue_items:
            LOGGER.info('Found %d new uuids, adding them to the queue' % len(queue_items))
            msgs = [Content(uuid) for uuid in queue_items]
            deferreds = [self.chan.basic_publish(exchange=self.amqp_exchange, content=msg) for msg in msgs]
            d = DeferredList(deferreds, consumeErrors=True)
            d.addCallbacks(self._addToQueueComplete, self._addToQueueErr)
        else:
            self.enqueueCallLater = reactor.callLater(1, self.enqueue)
        
    def _addToQueueComplete(self, data):
        LOGGER.info('Completed adding items into the queue...')
        self.enqueueCallLater = reactor.callLater(1, self.enqueue)
        
    def _addToQueueErr(self, error):
        LOGGER.error(error.printBriefTraceback)
        raise
            
    def remoteAddToHeap(self, uuid, type):
        LOGGER.debug('remoteAddToHeap: uuid=%s, type=%s' % (uuid, type))
        pass
        
    def createReservation(self, function_name, **kwargs):
        LOGGER.debug('%s Called' % function_name)
        if function_name == 'schedulerserver/remoteaddtoheap':
            if set(('uuid', 'type')).issubset(set(kwargs)):
                LOGGER.debug('\tUUID: %s\n\tType: %s' % (kwargs['uuid'], kwargs['type']))
                self.addToHeap(kwargs['uuid'], kwargs['type'])
                return {}
            else:
                return {'error': 'invalid parameters passed: required parameters are uuid and type'}
        return
        
    def addToHeap(self, uuid, type):
        # lookup if type is in the service_mapping, if it is
        # then rewrite type to the proper resource
        if self.service_mapping and self.service_mapping.has_key(type):
            LOGGER.info('Remapping resource %s to %s' % (type, self.service_mapping[type]))
            type = self.service_mapping[type]
        # Make sure the uuid is in bytes
        uuid = UUID(uuid).bytes
        interval = None
        if self.functions.has_key(type) and self.functions[type].has_key('interval'):
            interval = int(self.functions[type]['interval'])
        try:
            type = self.function_names.index(type)
        except:
            LOGGER.error("Unsupported function type: %s" % type)
            return
        if not interval and self.functions.has_key(type) and self.functions[type].has_key('interval'):
            interval = int(self.functions[type]['interval'])
        else:
            internal = 18000
        enqueue_time = int(time.time() + interval)
        # Add a UUID to the heap.
        LOGGER.debug('Adding %s to heap with enqueue_time %s and interval of %s' % (UUID(bytes=uuid).hex, enqueue_time, interval))
        heappush(self.heap, (enqueue_time, (uuid, interval)))


import cStringIO
import gzip as gzip_package
import base64
import hmac
import hashlib
import logging
import time
import urllib
import xml.etree.cElementTree as ET
from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList, deferredGenerator, waitForDeferred
from ..unicodeconverter import convertToUTF8
from ..requestqueuer import RequestQueuer
from .lib import return_true, etree_to_dict


S3_NAMESPACE = "{http://s3.amazonaws.com/doc/2006-03-01/}"
LOGGER = logging.getLogger("main")

class AmazonS3:
    """
    Amazon Simple Storage Service API.
    """
   
    ACCEPTABLE_ERROR_CODES = [400, 403, 404, 409]
    host = "s3.amazonaws.com"
    reserved_headers = ["x-amz-id-2", "x-amz-request-id", "date", "last-modified", "etag", "content-type", "content-length", "server"]
    
    def __init__(self, aws_access_key_id, aws_secret_access_key, rq=None):
        """
        **Arguments:**
         * *aws_access_key_id* -- Amazon AWS access key ID
         * *aws_secret_access_key* -- Amazon AWS secret access key
       
        **Keyword arguments:**
         * *rq* -- Optional RequestQueuer object.
        """
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
        self.rq.setHostMaxRequestsPerSecond(self.host, 0)
        self.rq.setHostMaxSimultaneousRequests(self.host, 0)
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
       
    def checkAndCreateBucket(self, bucket):
        """
        Check for a bucket's existence. If it does not exist, 
        create it.
       
        **Arguments:**
         * *bucket* -- Bucket name
        """
        d = self.getBucket(bucket)
        d.addErrback(self._checkAndCreateBucketErrback, bucket) 
        return d

    def _checkAndCreateBucketErrback(self, error, bucket):
        if int(error.value.status) == 404:    
            d = self.putBucket(bucket)
            d.addErrback(self._checkAndCreateBucketErrback2, bucket)
            return d
        raise Exception("Could not find or create bucket '%s'." % bucket_name)

    def _checkAndCreateBucketErrback2( self, error, bucket_name):
        raise Exception("Could not create bucket '%s'" % bucket_name)      
    
    def emptyBucket(self, bucket):
        """
        Delete all items in a bucket.
       
        **Arguments:**
         * *bucket* -- Bucket name
        """
        d = self.getBucket(bucket)
        d.addCallback(self._emptyBucketCallback, bucket)
        return d
       
    def _emptyBucketCallback(self, result, bucket): 
        xml = ET.XML(result["response"])
        key_nodes = xml.findall(".//%sKey" % S3_NAMESPACE)
        delete_deferreds = []
        for node in key_nodes:
            delete_deferreds.append(self.deleteObject(bucket, node.text))
        if len(delete_deferreds) == 0:
            return True
        d = DeferredList(delete_deferreds)
        if xml.find('.//%sIsTruncated' % S3_NAMESPACE).text == "false":
            d.addCallback(return_true)
        else:
            d.addCallback(self._emptyBucketCallbackRepeat, bucket)
        return d
   
    def _emptyBucketCallbackRepeat(self, data, bucket):
        return self.emptyBucket(bucket)
   
#    def listObjects(self, bucket, marker=None):
#        """
#        List information about the objects in the bucket.
#
#        **Arguments:**
#         * *bucket* -- Bucket name
#        """
#        bucket = convertToUTF8(bucket)
#        path = bucket
#        if marker is not None:
#            path = urllib.urlencode({"marker":marker})
#        headers = self._getAuthorization("GET", "", "", {}, "/" + path)
#        url = "http://%s/%s" % (self.host, path)
#        d = self.rq.getPage(url, method="GET", headers=headers)
#        d.addCallback(self._listObjectsCallback, bucket)
#        d.addErrback(self._genericErrback, url, method="GET", headers=headers)
#        return d
#    
#    @deferredGenerator
#    def _listObjectsCallback(self, result, bucket):
#        xml = ET.XML(result["response"])            
#        data = etree_to_dict(xml, namespace='{http://s3.amazonaws.com/doc/2006-03-01/}')
#        for obj in data["Contents"]:
#            for key in obj:
#                obj[key] = obj[key][0]
#            for key in obj["Owner"]:
#                obj["Owner"][key] = obj["Owner"][key][0]
#            yield obj
#        
#        if data["IsTruncated"][0] == "true":
#            d = self.listObjects(bucket, marker=obj["Key"])
#            wfd = waitForDeferred(d)
#            yield wfd
            
    def getBucket(self, bucket):
        """
        List information about the objects in the bucket.
       
        **Arguments:**
         * *bucket* -- Bucket name
        """
        bucket = convertToUTF8(bucket)
        headers = self._getAuthorization("GET", "", "", {}, "/" + bucket)
        url = "http://%s/%s" % (self.host, bucket)
        d = self.rq.getPage(url, method="GET", headers=headers)
        d.addErrback(self._genericErrback, url, method="GET", headers=headers)
        return d       
        
    def putBucket(self, bucket):
        """
        Creates a new bucket.
       
        **Arguments:**
         * *bucket* -- Bucket name
        """
        bucket = convertToUTF8(bucket)
        headers = {
            "Content-Length": 0
        }
        auth = self._getAuthorization("PUT", "", "", headers, "/" + bucket)
        headers.update(auth)
        url = "http://%s/%s" % (self.host, bucket)
        d = self.rq.getPage(url, method="PUT", headers=headers)
        d.addErrback(self._genericErrback, url, method="PUT", headers=headers)
        return d

    def deleteBucket(self, bucket):
        """
        Delete the bucket.
       
        **Arguments:**
         * *bucket* -- Bucket name
        """
        bucket = convertToUTF8(bucket)
        headers = {
            "Content-Length": 0
        }
        auth = self._getAuthorization("DELETE", "", "", headers, "/" + bucket)
        headers.update(auth)
        url = "http://%s/%s" % (self.host, bucket)
        d = self.rq.getPage(url, method="DELETE", headers=headers)
        d.addErrback(self._genericErrback, url, method="DELETE",    
                     headers=headers)
        return d

    def headObject(self, bucket, key):
        """
        Retrieve information about a specific object or object size, without
        actually fetching the object itself.
       
        **Arguments:**
         * *bucket* -- Bucket name
         * *key* -- Key name
        """
        bucket = convertToUTF8(bucket)
        key = convertToUTF8(key)
        path = "/" + bucket + "/" + key
        headers = self._getAuthorization("HEAD", "", "", {}, path)
        url = "http://%s/%s/%s" % (self.host, bucket, key)
        d = self.rq.getPage(url, method="HEAD", headers=headers)
        d.addCallback(self._getObjectCallback)
        d.addErrback(self._genericErrback, url, method="HEAD", headers=headers)
        return d

    def _decodeAmazonHeaders(self, headers):
        """
        Remove custom header prefix from header dictionary keys.
        """
        for key in headers.keys():
            if ("x-amz-meta-%s" % key.lower()) in self.reserved_headers:
                message = "Header %s is reserved for use by Amazon S3." % key
                LOGGER.critical(message)
                raise Exception(message)
        keys = headers.keys()
        values = headers.values()
        meta = "x-amz-meta-"
        return dict(zip([x.replace(meta,"") for x in keys], values))

    def getObject(self, bucket, key):
        """
        Returns object directly from Amazon S3 using a client/server 
        delivery mechanism.
       
        **Arguments:**
         * *bucket* -- Bucket name
         * *key* -- Key name
        """       
        bucket = convertToUTF8(bucket)
        key = convertToUTF8(key)
        path = "/" + bucket + "/" + key
        headers = self._getAuthorization("GET", "", "", {}, path)
        url = "http://%s/%s/%s" % (self.host, bucket, key)
        d = self.rq.getPage(url, method="GET", headers=headers)
        d.addCallback(self._getObjectCallback)
        d.addErrback(self._genericErrback, url, method="GET", headers=headers)
        return d   
       
    def _getObjectCallback(self, data):
        if "content-encoding" in data["headers"]:
            if "gzip" in data["headers"]["content-encoding"]:
                compressedstream = cStringIO.StringIO(data["response"])
                zfile = gzip_package.GzipFile(fileobj=compressedstream)
                data["response"] = zfile.read()
        data["headers"] = self._decodeAmazonHeaders(data["headers"])
        return data
   
    def _encodeAmazonHeaders(self, headers):
        """
        Prepend custom header prefix to header dictionary keys.
        """
        headers = dict([(x[0].lower(), x[1]) for x in headers.items()])
        for header in self.reserved_headers:
            if header in headers:
                del headers[header]
        keys = headers.keys()
        values = headers.values()
        #for key in keys:
        #    if key.lower() in self.reserved_headers:
        #        message = "Header %s is reserved for use by Amazon S3." % key
        #        LOGGER.error(message)
        #        raise Exception(message)
        meta = "x-amz-meta-"
        return dict(zip(["%s%s" % (meta, x) for x in keys], values))
       
    def putObject(self, bucket, key, data, content_type="text/html", 
                  public=True, headers=None, gzip=False):
        """
        Add an object to a bucket.
       
        **Arguments:**
         * *bucket* -- Bucket name
         * *key* -- Key name
         * *data* -- Data string
       
        **Keyword arguments:**
         * *content_type* -- Content type header (Default 'text/html')
         * *public* -- Boolean flag representing access (Default True)
         * *headers* -- Custom header dictionary (Default empty dictionary)
         * *gzip* -- Boolean flag to gzip data (Default False)
        """
        bucket = convertToUTF8(bucket)
        key = convertToUTF8(key)
        if not isinstance(data, str):
            data = convertToUTF8(data)
        if headers is None:
            headers = {}
        # Wrap user-defined headers in with the Amazon custom header prefix.
        headers = self._encodeAmazonHeaders(headers)
        if gzip:
            # Gzip that bastard!
            headers["content-encoding"] = "gzip"
            zbuf = cStringIO.StringIO()
            zfile = gzip_package.GzipFile(None, 'wb', 9, zbuf)
            zfile.write(data)
            zfile.close()
            data = zbuf.getvalue()
        content_md5 = base64.encodestring(hashlib.md5(data).digest()).strip()
        if public:
            headers['x-amz-acl'] = 'public-read'
        else:
            headers['x-amz-acl'] = 'private'
        headers.update({
            'Content-Length':len(data),
            'Content-Type':content_type,
            'Content-MD5':content_md5
        })
        path = "/" + bucket + "/" + key
        auth = self._getAuthorization("PUT", content_md5, content_type, 
                                      headers, path)
        headers.update(auth)
        url = "http://%s/%s/%s" % (self.host, bucket, key)
        d = self.rq.getPage(url, method="PUT", headers=headers, postdata=data)
        d.addErrback(self._genericErrback, url, method="PUT", headers=headers, 
                     postdata=data)
        return d

    def deleteObject(self, bucket, key):
        """
        Remove the specified object from Amazon S3.
       
        **Arguments:**
         * *bucket* -- Bucket name
         * *key* -- Key name
        """
        bucket = convertToUTF8(bucket)
        key = convertToUTF8(key)
        path = "/" + bucket + "/" + key
        headers = self._getAuthorization("DELETE", "", "", {}, path)
        url = "http://%s/%s/%s" % (self.host, bucket, key)
        d = self.rq.getPage(url, method="DELETE", headers=headers)
        d.addErrback(self._genericErrback, url, method="DELETE", 
                     headers=headers)
        return d
           
    def _genericErrback(self, error, url, method="GET", headers=None,
                        postdata=None, count=0):
        if headers is None:
            headers = {}
        if "status" in error.value.__dict__:
            # 204, empty response but otherwise OK, as in the case of a delete.
            # Move on, nothing to see here.
            if int(error.value.status) == 204:
                return {
                        "response":error.value.response,
                        "status":int(error.value.status),
                        "headers":error.value.response,
                        "message":error.value.headers
                    }
            # Something other than a 40x error. Something is wrong, but let's
            # try that again a few times.          
            elif int(error.value.status) not in self.ACCEPTABLE_ERROR_CODES \
                 and count < 3:
                d = self.rq.getPage(url,       
                                    method=method,
                                    headers=headers,
                                    postdata=postdata
                                    )
                                   
                d.addErrback(self._genericErrback,
                                url,
                                method=method,
                                headers=headers,
                                postdata=postdata,
                                count=count + 1
                                )
                return d
            # 404 or other normal error, pass it along.
            else:
                return error   
        else:
            return error
               
    def _canonicalize(self, headers):
        """
        Canonicalize headers for use with AWS Authorization.
       
        **Arguments:**
         * *headers* -- Key value pairs of headers
       
        **Returns:**
         * A string representation of the parameters.
        """
        keys = [k for k in headers.keys() if k.startswith("x-amz-")]
        keys.sort(key = str.lower)
        return "\n".join(["%s:%s" % (x, headers[x]) for x in keys])

    def _getAuthorization(self, method, content_hash, content_type, 
                          headers, resource):
        """
        Create authentication headers.
       
        **Arguments:**
         * *method* -- HTTP method of the request
         * *parameters* -- Key value pairs of parameters
       
        **Returns:**
         * A dictionary of authorization parameters
        """
        date = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime())
        amazon_headers = self._canonicalize(headers)
        if len(amazon_headers):
            data = "%s\n%s\n%s\n%s\n%s\n%s" % (
                method,
                content_hash,
                content_type,
                date,
                amazon_headers,
                resource
           )
        else:
            data = "%s\n%s\n%s\n%s\n%s%s" % (
                method,
                content_hash,
                content_type,
                date,
                amazon_headers,
                resource
           )
        args = [self.aws_secret_access_key, data, hashlib.sha1]
        signature = base64.encodestring(hmac.new(*args).digest()).strip()
        authorization = "AWS %s:%s" % (self.aws_access_key_id, signature)
        return {'Authorization': authorization, 'Date':date, "Host":self.host}




import base64
import hmac
import hashlib
import urllib
import xml.etree.cElementTree as ET
from datetime import datetime, timedelta
import simplejson
from ..unicodeconverter import convertToUTF8
from ..requestqueuer import RequestQueuer
from .lib import return_true, safe_quote_tuple


SQS_NAMESPACE = "{http://queue.amazonaws.com/doc/2009-02-01/}"


class AmazonSQS:
   
    """
    Amazon Simple Queue Service API.
    """
   
    host = "queue.amazonaws.com"
   
    def __init__(self, aws_access_key_id, aws_secret_access_key, rq=None):
        """
        **Arguments:**
         * *aws_access_key_id* -- Amazon AWS access key ID string
         * *aws_secret_access_key* -- Amazon AWS secret access key string
       
        **Keyword arguments:**
         * *rq* -- Optional RequestQueuer object.
        """
        if rq is None:
            self.rq = RequestQueuer()
        else:
            self.rq = rq
        self.rq.setHostMaxRequestsPerSecond(self.host, 0)
        self.rq.setHostMaxSimultaneousRequests(self.host, 0)
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key  

    def listQueues(self, name_prefix=None):
        """
        Return a list of your queues. The maximum number of queues
        that can be returned is 1000. If you specify a value for
        the optional name_prefix parameter, only queues with
        a name beginning with the specified value are returned.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryListQueues.html
       
        **Keyword arguments:**
         * *name_prefix* -- String to use for filtering the list results.
           Only those queues whose name begins with the specified
           string are returned. (Default None)

        ****Returns:****
         * Deferred of a list of queue paths that can be used as the
           'resource' argument for other class methods.
        """
        parameters = {
            "Action":"ListQueues"}
        if name_prefix is not None:
            parameters["QueueNamePrefix"] = name_prefix
        d = self._request(parameters)
        d.addCallback(self._listQueuesCallback)
        return d

    def _listQueuesCallback(self, data):
        xml = ET.fromstring(data["response"])
        queue_urls = xml.findall(".//%sQueueUrl" % SQS_NAMESPACE)
        host_string = "https://%s" % self.host
        queue_paths = [x.text.replace(host_string, "") for x in queue_urls]
        return queue_paths

    def createQueue(self, name, visibility_timeout=None):
        """
        Create a new queue.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryCreateQueue.html
       
        **Arguments:**
         * *name* -- The name to use for the queue created.
       
        **Keyword arguments:**
         * *visibility_timeout* -- The visibility timeout (in seconds) to use
           for this queue. (Default None)

        **Returns:**
         * Deferred of a queue path string that can be used as the 'resource' argument
           for other class methods.
        """
        name = convertToUTF8(name)
        parameters = {
            "Action":"CreateQueue",
            "QueueName":name}
        if visibility_timeout is not None:
            parameters["DefaultVisibilityTimeout"] = visibility_timeout
        d = self._request(parameters)
        d.addCallback(self._createQueueCallback)
        d.addErrback(self._genericErrback)
        return d
       
    def _createQueueCallback(self, data):
        xml = ET.fromstring(data["response"])
        queue_url = xml.find(".//%sQueueUrl" % SQS_NAMESPACE).text
        queue_path = queue_url.replace("https://%s" % self.host, "")
        return queue_path
       
    def deleteQueue(self, resource):
        """
        Delete the queue specified by resource, regardless of whether the
        queue is empty. If the specified queue does not exist, returns
        a successful response.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryDeleteQueue.html
       
        **Arguments:**
         * *resource* -- The path of the queue.

        **Returns:**
         * Deferred of boolean True.
        """
        parameters = {"Action":"DeleteQueue"}
        d = self._request(parameters, resource=resource, method="DELETE")
        d.addCallback(return_true)
        d.addErrback(self._genericErrback)
        return d
       
    def setQueueAttributes(self, resource, visibility_timeout=None,
                           policy=None):
        """
        Set attributes of a queue.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QuerySetQueueAttributes.html
       
        **Keyword arguments:**
         * *visibility_timeout* -- The visibility timeout (in seconds) to use
           for this queue. (Default None)
         * *policy* -- Python object representing a valid policy. (Default None)
           See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?AccessPolicyLanguage_Concepts.html
       
        **Arguments:**
         * *resource* -- The path of the queue.

        **Returns:**
         * Deferred of boolean True.
        """
        parameters = {"Action":"SetQueueAttributes"}
        attributes = {}
        if policy is not None:
            attributes["Policy"] = simplejson.dumps(policy)
        if visibility_timeout is not None:
            attributes["VisibilityTimeout"] = visibility_timeout
        attr_count = 1
        for name in attributes:
            parameters["Attribute.%s.Name" % attr_count] = name
            parameters["Attribute.%s.Value" % attr_count] = attributes[name]
            attr_count += 1
        d = self._request(parameters, resource=resource)
        d.addCallback(return_true)
        d.addErrback(self._genericErrback)
        return d
   
    def getQueueAttributes(self, resource, name=None):
        """
        Get attributes of a queue.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryGetQueueAttributes.html
       
        **Keyword arguments:**
         * *name* -- The attribute you want to get. (Default All.) Valid 
           values are ``(All | ApproximateNumberOfMessages | VisibilityTimeout
           | CreatedTimestamp | LastModifiedTimestamp | Policy)``
       
        **Arguments:**
         * *resource* -- The path of the queue.

        **Returns:**
         * Deferred of a dictionary of the requested attributes.
        """
        attributes = [
            "All",
            "ApproximateNumberOfMessages",
            "VisibilityTimeout",
            "CreatedTimestamp",
            "LastModifiedTimestamp",
            "Policy"]
        if name is not None:
            if name not in attributes:
                raise Exception("Unknown attribute name '%s'." % name)
        else:
            name = "All"
        parameters = {
            "Action":"GetQueueAttributes",
            "AttributeName":name}       
        d = self._request(parameters, resource=resource, method="DELETE")
        d.addCallback(self._getQueueAttributesCallback)
        d.addErrback(self._genericErrback)
        return d
           
    def _getQueueAttributesCallback(self, data):
        attributes = {}
        xml = ET.fromstring(data["response"])
        xml_attributes = xml.findall(".//%sAttribute" % SQS_NAMESPACE)
        for attribute in xml_attributes:
            name = attribute.find(".//%sName" % SQS_NAMESPACE).text
            value = attribute.find(".//%sValue" % SQS_NAMESPACE).text
            attributes[name] = value
        if "Policy" in attributes:
            attributes["Policy"] = simplejson.loads(attributes["Policy"])
        integer_attribute_names = [
            'CreatedTimestamp',
            'ApproximateNumberOfMessages',
            'LastModifiedTimestamp',
            'VisibilityTimeout']
        for name in integer_attribute_names:
            if name in attributes:
                attributes[name] = int(attributes[name])
        return attributes
           
    def addPermission(self, resource, label, aws_account_ids, actions=None):
        """
        Add a permission to a queue for a specific principal. This allows
        for sharing access to the queue.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryAddPermission.html
       
        **Arguments:**
         * *resource* -- The path of the queue.
         * *label* -- The unique identification string of the permission you're
           setting.
         * *aws_account_ids* -- A string or list of strings of (a) valid 
           12-digit AWS account number(s), with or without hyphens.
       
        **Keyword arguments:**
         * *actions* -- A string or list of strings of actions you want to
           allow for the specified principal. Default '*'. Valid values are (* |
           SendMessage | ReceiveMessage | DeleteMessage | ChangeMessageVisibility
           | GetQueueAttributes)
       
        **Returns:**
         * Deferred of boolean True.
        """
        if actions is None:
            actions = ["*"]
        if isinstance(actions, str) or isinstance(actions, unicode):
            actions = [str(actions)]
        if not isinstance(actions, list):
            raise Exception("Actions must be a string or list of strings.")
            
        if isinstance(aws_account_ids, str) or \
           isinstance(aws_account_ids, unicode):
            aws_account_ids = [str(aws_account_ids)]
        if not isinstance(aws_account_ids, list):
            message = "aws_account_ids must be a string or list of strings."
            raise Exception(message)   
        aws_account_ids = [x.replace("-", "") for x in aws_account_ids]   
        action_options = [
            "*",
            "SendMessage",
            "ReceiveMessage",
            "DeleteMessage",
            "ChangeMessageVisibility",
            "GetQueueAttributes"]
        for action in actions:
            if action not in action_options:
                raise Exception("Unknown action name '%s'." % action)
        if len(actions) == 0:
            actions.append("*")
        parameters = {
            "Action":"AddPermission",
            "Label":label}
        action_count = 1
        for name in actions:
            parameters["ActionName.%s" % action_count] = name
            action_count += 1
        aws_account_id_count = 1
        for name in aws_account_ids:
            parameters["AWSAccountId.%s" % aws_account_id_count] = name
            aws_account_id_count += 1
        d = self._request(parameters, resource=resource)
        d.addCallback(return_true)
        d.addErrback(self._genericErrback)
        return d 

    def removePermission(self, resource, label):
        """
        Revokes any permissions in the queue policy that matches the
        label parameter.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryRemovePermission.html
       
        **Arguments:**
         * *resource -- The path of the queue.
         * *label -- The identfication of the permission you want to remove.
       
        **Returns:**
         * Deferred of boolean True.
        """
        parameters = {
            "Action":"RemovePermission",
            "Label":label}
        d = self._request(parameters, resource=resource)
        d.addCallback(return_true)
        d.addErrback(self._genericErrback)
        return d 

    def sendMessage(self, resource, message):
        """
        Deliver a message to the specified queue. The maximum allowed message
        size is 8 KB.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QuerySendMessage.html
       
        **Arguments:**
         * *resource* -- The path of the queue.
         * *message* -- The message to send. Characters must be in: ( #x9 | #xA |
           #xD | [#x20 to #xD7FF] | [#xE000 to #xFFFD] | [#x10000 to #x10FFFF] )
           See: http://www.w3.org/TR/REC-xml/#charsets
       
        **Returns:**
         * Deferred of a string with a Message ID.
        """
        parameters = {
            "Action":"SendMessage",
            "MessageBody":message}
        d = self._request(parameters, resource=resource)
        d.addCallback(self._sendMessageCallback)
        d.addErrback(self._genericErrback)
        return d 
   
    def _sendMessageCallback(self, data):
        xml = ET.fromstring(data["response"])
        return xml.find(".//%sMessageId" % SQS_NAMESPACE).text

    def receiveMessage(self, resource, visibility_timeout=None,
                       max_number_of_messages=1, get_sender_id=False,
                       get_sent_timestamp=False):
        """
        Retrieve one or more messages from the specified queue.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryReceiveMessage.html
       
        **Arguments:**
         * *resource* -- The path of the queue.
       
        **Keyword arguments:**
         * *visibility_timeout* -- The duration (in seconds) that the received
           messages are hidden from subsequent retrieve requests after being
           retrieved by a ReceiveMessage request. (Default None)
         * *max_number_of_messages* -- Maximum number of messages to return. 1-10.
           (Default 1)
         * *get_sender_id* -- Retrieve the sender id attribute. (Default False)
         * *get_sent_timestamp* -- Retrieve the timestamp attribute. (Default False)
       
        **Returns:**
         * Deferred of dictionary with attributes 'md5', 'id', 'receipt_handle',
           'body'. Optionally 'sender_id' and 'sent_timestamp'.
        """
        parameters = {
            "Action":"ReceiveMessage"}
        if visibility_timeout is not None:
            parameters["VisibilityTimeout"] = int(visibility_timeout)
        if int(max_number_of_messages) > 1:
            parameters["MaxNumberOfMessages"] = int(max_number_of_messages)
        attr_count = 1
        if get_sender_id == True:
            parameters["AttributeName.%s" % attr_count] = "SenderId"
            attr_count += 1
        if get_sent_timestamp == True:
            parameters["AttributeName.%s" % attr_count] = "SentTimestamp"
            attr_count += 1
        d = self._request(parameters, resource=resource)
        d.addCallback(self._receiveMessageCallback)
        d.addErrback(self._genericErrback)
        return d
       
    def _receiveMessageCallback(self, data):
        messages = []
        xml = ET.fromstring(data["response"])
        message_elements = xml.findall(".//%sMessage" % SQS_NAMESPACE)
        for element in message_elements:
            message = {}
            attributes = element.findall(".//%sAttributes" % SQS_NAMESPACE)
            for attribute in attributes:
                name = attribute.find(".//%sName" % SQS_NAMESPACE).text
                value = attribute.find(".//%sValue" % SQS_NAMESPACE).text
                if name == "SenderId":
                    message['sender_id'] = value
                elif name == "SentTimestamp":
                    message['sent_timestamp'] = value
            message["md5"] = element.find(".//%sMD5OfBody" % SQS_NAMESPACE).text
            message["id"] = element.find(".//%sMessageId" % SQS_NAMESPACE).text
            message["receipt_handle"] = element.find(".//%sReceiptHandle" % SQS_NAMESPACE).text
            message["body"] = element.find(".//%sBody" % SQS_NAMESPACE).text
            messages.append(message)
        return messages

    def deleteMessage(self, resource, receipt_handle):
        """
        Delete the specified message from the specified queue.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryDeleteMessage.html
       
        **Arguments:**
         * *resource* -- The path of the queue.
         * *receipt_handle* -- The receipt handle associated with the message
           you want to delete.
       
        **Returns:**
         * Deferred of boolean True.
        """
        parameters = {
            "Action":"DeleteMessage",
            "ReceiptHandle":receipt_handle}
        d = self._request(parameters, resource=resource)
        d.addCallback(return_true)
        d.addErrback(self._genericErrback)
        return d
   
    def changeMessageVisibility(self, resource, receipt_handle, 
                                visibility_timeout):
        """
        Change the visibility timeout of a specified message in a queue to
        a new value.
        
        See: http://docs.amazonwebservices.com/AWSSimpleQueueService/2009-02-01/SQSDeveloperGuide/index.html?Query_QueryChangeMessageVisibility.html
       
        **Arguments:**
         * *resource* -- The path of the queue.
         * *receipt_handle* -- The receipt handle associated with the message
           you want to delete.
         * *visibility_timeout* -- The new value for the message's visibility timeout (in seconds).
       
        **Returns:**
         * Deferred of boolean True.
        """
        parameters = {
            "Action":"ChangeMessageVisibility",
            "ReceiptHandle":receipt_handle,
            "VisibilityTimeout":int(visibility_timeout)}
        d = self._request(parameters, resource=resource)
        d.addCallback(return_true)
        d.addErrback(self._genericErrback)
        return d
      
    def _genericErrback(self, error):
        """
        Use server specified error when possible.
       
        **Arguments:**
         * *error* -- Twisted error
       
        **Returns:**
         * Twisted error
        """
        if hasattr(error, "value"):
            if hasattr(error.value, "response"):
                xml = ET.fromstring(error.value.response)
                message = xml.find(".//%sMessage" % SQS_NAMESPACE).text
                raise Exception(message)
        return error       
       
    def _canonicalize(self, parameters):
        """
        Canonicalize parameters for use with AWS Authorization.
       
        **Arguments:**
         * *parameters* -- Key value pairs of parameters
       
        **Returns:**
         * A safe-quoted string representation of the parameters.
        """
        parameters = parameters.items()
        # Alphebetize key-value pairs.
        parameters.sort(lambda x, y:cmp(x[0], y[0]))
        # Safe-quote and combine parameters into a string
        return "&".join([safe_quote_tuple(x) for x in parameters])

    def _request(self, parameters, method="GET", resource="/"):
        """
        Add authentication parameters and make request to Amazon.
       
        **Arguments:**
         * *parameters* -- Key value pairs of parameters
       
        **Keyword arguments:**
         * *method* -- HTTP request method
         * *resource* -- Requested server resource
       
        **Returns:**
         * Deferred of server response.
        """

        parameters = self._getAuthorization(method, parameters, 
                                            resource=resource)
           
        query_string = urllib.urlencode(parameters)

        url = "https://%s%s?%s" % (self.host, resource, query_string)
       
        #print url
       
        d = self.rq.getPage(url, method=method)
        return d

    def _getAuthorization(self, method, parameters, resource="/"):
        """
        Create authentication parameters.
       
        **Arguments:**
         * *method* -- HTTP method of the request
         * *parameters* -- Key value pairs of parameters
       
        **Returns:**
         * A dictionary of authorization parameters
        """
        expires = datetime.utcnow() + timedelta(30)
        signature_parameters = {
            "AWSAccessKeyId":self.aws_access_key_id,
            "SignatureVersion":"2",
            "SignatureMethod":"HmacSHA256",
            'Expires':"%s+00:00" % expires.isoformat()[0:19],
            "AWSAccessKeyId":self.aws_access_key_id,
            "Version":"2009-02-01"}
        signature_parameters.update(parameters)
        query_string = self._canonicalize(signature_parameters)
        string_to_sign = "%(method)s\n%(host)s\n%(resource)s\n%(qs)s" % {
            "method":method,
            "host":self.host.lower(),
            "resource":resource,
            "qs":query_string}
        args = [self.aws_secret_access_key, string_to_sign, hashlib.sha256]
        signature = base64.encodestring(hmac.new(*args).digest()).strip()
        signature_parameters.update({'Signature':signature})
        return signature_parameters



from .base import BaseServer, LOGGER
from ..resources2 import WorkerResource
from ..networkaddress import getNetworkAddress
from ..amqp import amqp as AMQP
from ..resources import InterfaceResource, ExposedResource
from MySQLdb.cursors import DictCursor
from twisted.internet import reactor, protocol, task
from twisted.enterprise import adbapi
from twisted.web import server
from twisted.protocols.memcache import MemCacheProtocol, DEFAULT_PORT
from twisted.internet.defer import Deferred, DeferredList, maybeDeferred, inlineCallbacks
from twisted.internet.threads import deferToThread
from uuid import UUID, uuid4
import pprint
import simplejson

PRETTYPRINTER = pprint.PrettyPrinter(indent=4)

class WorkerServer(BaseServer):
    
    public_ip = None
    local_ip = None
    network_information = {}
    simultaneous_jobs = 100
    doSomethingCallLater = None
    jobs_complete = 0
    job_queue = []
    job_queue_a = job_queue.append
    
    def __init__(self,
            aws_access_key_id,
            aws_secret_access_key,
            aws_s3_http_cache_bucket=None,
            aws_s3_storage_bucket=None,
            mysql_username=None,
            mysql_password=None,
            mysql_host=None,
            mysql_database=None,
            amqp_host=None,
            amqp_username=None,
            amqp_password=None,
            amqp_vhost=None,
            amqp_queue=None,
            amqp_exchange=None,
            memcached_host=None,
            service_mapping=None,
            service_args_mapping=None,
            amqp_port=5672,
            amqp_prefetch_count=1000,
            mysql_port=3306,
            memcached_port=11211,
            max_simultaneous_requests=100,
            max_requests_per_host_per_second=0,
            max_simultaneous_requests_per_host=0,
            port=5005,
            log_file='workerserver.log',
            log_directory=None,
            log_level="debug"):
        self.network_information["port"] = port
        # Create MySQL connection.
        self.mysql = adbapi.ConnectionPool(
            "MySQLdb",
            db=mysql_database,
            port=mysql_port,
            user=mysql_username,
            passwd=mysql_password,
            host=mysql_host,
            cp_reconnect=True,
            cursorclass=DictCursor)
        # Create Memcached client
        self.memcached_host = memcached_host
        self.memcached_port = memcached_port
        self.memc_ClientCreator = protocol.ClientCreator(
            reactor, MemCacheProtocol)
        # Resource Mappings
        self.service_mapping = service_mapping
        self.service_args_mapping = service_args_mapping
        # HTTP interface
        resource = WorkerResource(self)
        self.site_port = reactor.listenTCP(port, server.Site(resource))
        # Create AMQP Connection
        # AMQP connection parameters
        self.amqp_host = amqp_host
        self.amqp_vhost = amqp_vhost
        self.amqp_port = amqp_port
        self.amqp_username = amqp_username
        self.amqp_password = amqp_password
        self.amqp_queue = amqp_queue
        self.amqp_exchange = amqp_exchange
        self.amqp_prefetch_count = amqp_prefetch_count
        BaseServer.__init__(
            self,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_s3_http_cache_bucket=aws_s3_http_cache_bucket,
            aws_s3_storage_bucket=aws_s3_storage_bucket,
            max_simultaneous_requests=max_simultaneous_requests,
            max_requests_per_host_per_second=max_requests_per_host_per_second,
            max_simultaneous_requests_per_host=max_simultaneous_requests_per_host,
            log_file=log_file,
            log_directory=log_directory,
            log_level=log_level)
    
    def start(self):
        reactor.callWhenRunning(self._start)
        return self.start_deferred
    
    @inlineCallbacks
    def _start(self):
        yield self.getNetworkAddress()
        # Create memcached client
        self.memc = yield self.memc_ClientCreator.connectTCP(self.memcached_host, self.memcached_port)
        LOGGER.info('Connecting to broker.')
        self.conn = yield AMQP.createClient(
            self.amqp_host,
            self.amqp_vhost,
            self.amqp_port)
        self.auth = yield self.conn.authenticate(
            self.amqp_username,
            self.amqp_password)
        self.chan = yield self.conn.channel(1)
        yield self.chan.channel_open()
        yield self.chan.basic_qos(prefetch_count=self.amqp_prefetch_count)
        # Create Queue
        yield self.chan.queue_declare(
            queue=self.amqp_queue,
            durable=False,
            exclusive=False,
            auto_delete=False)
        yield self.chan.queue_bind(
            queue=self.amqp_queue,
            exchange=self.amqp_exchange)
        d = self.chan.basic_consume(queue=self.amqp_queue,
            no_ack=False,
            consumer_tag="awspider_consumer")
        d.addCallback(self.dequeue)
        self.queue = yield self.conn.queue("awspider_consumer")
        yield BaseServer.start(self)
        self.jobsloop = task.LoopingCall(self.executeJobs)
        self.jobsloop.start(1)
    
    @inlineCallbacks
    def shutdown(self):
        LOGGER.debug("Closting connection")
        try:
            self.doSomethingCallLater.cancel()
        except:
            pass
        # Shut things down
        LOGGER.info('Closing broker connection')
        yield self.chan.channel_close()
        chan0 = yield self.conn.channel(0)
        yield chan0.connection_close()
        
    def dequeue(self, consumer=None):
        LOGGER.info('Starting Dequeuing Thread...')
        deferToThread(self._dequeue)
    
    def _dequeue(self):
        if len(self.active_jobs) <= self.amqp_prefetch_count:
            d = self.queue.get()
            d.addCallback(self._dequeueCallback)
            d.addErrback(self.workerError)
        else:
            reactor.callLater(1, self._dequeue)
    
    def _dequeueCallback(self, msg):
        # Get the hex version of the UUID from byte string we were sent
        uuid = UUID(bytes=msg.content.body).hex
        d = self.getJob(uuid, msg.delivery_tag)
        d.addCallback(self._dequeueCallback2, msg)
        d.addErrback(self.workerError)
    
    def _dequeueCallback2(self, job, msg):
        if job:
            # Load custom function.
            if job['function_name'] in self.functions:
                job['exposed_function'] = self.functions[job['function_name']]
            else:
                LOGGER.error("Could not find function %s." % job['function_name'])
            LOGGER.info('Pulled job off of AMQP queue')
            job['kwargs'] = self.mapKwargs(job)
            job['delivery_tag'] = msg.delivery_tag
            self.job_queue_a(job)
        self._dequeue()
    
    def executeJobs(self):
        while len(self.job_queue) > 0 and len(self.active_jobs) < self.simultaneous_jobs:
            job = self.job_queue.pop(0)
            exposed_function = job["exposed_function"]
            kwargs = job["kwargs"]
            function_name = job["function_name"]
            uuid = job["uuid"]
            d = self.callExposedFunction(
                exposed_function["function"], 
                kwargs, 
                function_name, 
                uuid=uuid)
            d.addCallback(self._executeJobCallback, job)
            d.addErrback(self.workerError)
        
    def _executeJobCallback(self, data, job):
        self.chan.basic_ack(delivery_tag=job['delivery_tag'])
        self.jobs_complete += 1
        LOGGER.info('Completed %d jobs' % self.jobs_complete)
        
    def workerError(self, error):
        LOGGER.error('Worker Error: %s' % str(error))
    
    def getJob(self, uuid, delivery_tag):
        d = self.memc.get(uuid)
        d.addCallback(self._getJobCallback, uuid, delivery_tag)
        d.addErrback(self.workerError)
        return d
    
    def _getJobCallback(self, account, uuid, delivery_tag):
        job = account[1]
        if not job:
            LOGGER.debug('Could not find uuid in memcached: %s' % uuid)
            sql = "SELECT account_id, type FROM spider_service WHERE uuid = '%s'" % uuid
            d = self.mysql.runQuery(sql)
            d.addCallback(self.getAccountMySQL, uuid, delivery_tag)
            d.addErrback(self.workerError)
            return d
        else:
            LOGGER.debug('Found uuid in memcached: %s' % uuid)
            return simplejson.loads(job)
    
    def getAccountMySQL(self, spider_info, uuid, delivery_tag):
        if spider_info:
            account_type = spider_info[0]['type'].split('/')[0]
            sql = "SELECT * FROM content_%saccount WHERE account_id = %d" % (account_type, spider_info[0]['account_id'])
            d = self.mysql.runQuery(sql)
            d.addCallback(self.createJob, spider_info, uuid)
            d.addErrback(self.workerError)
            return d
        LOGGER.info('No spider_info given for uuid %s' % uuid)
        self.chan.basic_ack(delivery_tag=delivery_tag)
        return None
    
    def createJob(self, account_info, spider_info, uuid):
        job = {}
        account = account_info[0]
        function_name = spider_info[0]['type']
        if self.service_mapping and self.service_mapping.has_key(function_name):
            LOGGER.info('Remapping resource %s to %s' % (function_name, self.service_mapping[function_name]))
            function_name = self.service_mapping[function_name]
        job['function_name'] = function_name
        # job['reservation_cache'] = None
        job['uuid'] = uuid
        job['account'] = account
        # Save account info in memcached for up to 7 days
        d = self.memc.set(uuid, simplejson.dumps(job), 60*60*24*7)
        d.addCallback(self._createJobCallback, job)
        d.addErrback(self.workerError)
        return d
    
    def _createJobCallback(self, memc, job):
        return job
        
    def mapKwargs(self, job):
        kwargs = {}
        service_name = job['function_name'].split('/')[0]
        # remap some fields that differ from the plugin and the database
        if service_name in self.service_args_mapping:
            for key in self.service_args_mapping[service_name]:
                if key in job['account']:
                    job['account'][self.service_args_mapping[service_name][key]] = job['account'][key]
        # apply job fields to req and optional kwargs
        for arg in job['exposed_function']['required_arguments']:
            if arg in job['account']:
                kwargs[arg] = job['account'][arg]
        for arg in job['exposed_function']['optional_arguments']:
            if arg in job['account']:
                kwargs[arg] = job['account'][arg]
        LOGGER.debug('Function: %s\nKWARGS: %s' %(job['function_name'], repr(kwargs)))
        return kwargs
        
    def getNetworkAddress(self):
        d = getNetworkAddress()
        d.addCallback(self._getNetworkAddressCallback)
        d.addErrback(self._getNetworkAddressErrback)
        return d
    
    def _getNetworkAddressCallback(self, data):
        if "public_ip" in data:
            self.public_ip = data["public_ip"]
            self.network_information["public_ip"] = self.public_ip
        if "local_ip" in data:
            self.local_ip = data["local_ip"]
            self.network_information["local_ip"] = self.local_ip
    
    def _getNetworkAddressErrback(self, error):
        message = "Could not get network address."
        LOGGER.error(message)
        raise Exception(message)


import pprint
from uuid import uuid4
from twisted.internet.defer import Deferred, DeferredList, maybeDeferred
from twisted.web.resource import Resource
from twisted.internet import reactor
from twisted.web import server
from .base import BaseServer, LOGGER
from ..resources import InterfaceResource, ExposedResource
from ..aws import sdb_now
from ..evaluateboolean import evaluateBoolean

PRETTYPRINTER = pprint.PrettyPrinter(indent=4)

class InterfaceServer(BaseServer):
    
    exposed_functions = []
    exposed_function_resources = {}
    
    def __init__(self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_sdb_reservation_domain, 
            aws_s3_reservation_cache_bucket=None,
            aws_s3_http_cache_bucket=None,
            aws_s3_storage_bucket=None,
            aws_sdb_coordination_domain=None,
            max_simultaneous_requests=50,
            max_requests_per_host_per_second=1,
            max_simultaneous_requests_per_host=5,
            port=5000, 
            log_file='interfaceserver.log',
            log_directory=None,
            log_level="debug",
            name=None,
            time_offset=None):
        if name == None:
            name = "AWSpider Interface Server UUID: %s" % self.uuid
        resource = Resource()
        interface_resource = InterfaceResource(self)
        resource.putChild("interface", interface_resource)
        self.function_resource = Resource()
        resource.putChild("function", self.function_resource)
        self.site_port = reactor.listenTCP(port, server.Site(resource))
        BaseServer.__init__(
            self,
            aws_access_key_id, 
            aws_secret_access_key, 
            aws_s3_reservation_cache_bucket=aws_s3_reservation_cache_bucket,
            aws_s3_http_cache_bucket=aws_s3_http_cache_bucket, 
            aws_sdb_reservation_domain=aws_sdb_reservation_domain, 
            aws_s3_storage_bucket=aws_s3_storage_bucket,
            aws_sdb_coordination_domain=aws_sdb_coordination_domain,
            max_simultaneous_requests=max_simultaneous_requests,
            max_requests_per_host_per_second=max_requests_per_host_per_second,
            max_simultaneous_requests_per_host=max_simultaneous_requests_per_host,
            log_file=log_file,
            log_directory=log_directory,
            log_level=log_level,
            name=name,
            time_offset=time_offset,
            port=port)
        
    def start(self):
        reactor.callWhenRunning(self._start)
        return self.start_deferred

    def _start(self):
        deferreds = []
        if self.time_offset is None:
            deferreds.append(self.getTimeOffset())
        d = DeferredList(deferreds, consumeErrors=True)
        d.addCallback(self._startCallback)

    def _startCallback(self, data):
        for row in data:
            if row[0] == False:
                d = self.shutdown()
                d.addCallback(self._startHandleError, row[1])
                return d
        d = BaseServer.start(self)

    def shutdown(self):
        deferreds = []
        LOGGER.debug("%s stopping on main HTTP interface." % self.name)
        d = self.site_port.stopListening()
        if isinstance(d, Deferred):
            deferreds.append(d)
        if len(deferreds) > 0:
            d = DeferredList(deferreds)
            d.addCallback(self._shutdownCallback)
            return d
        else:
            return self._shutdownCallback(None)

    def _shutdownCallback(self, data):
        return BaseServer.shutdown(self)
    
    def makeCallable(self, func, interval=0, name=None, expose=False):
        function_name = BaseServer.makeCallable(
            self, 
            func, 
            interval=interval, 
            name=name, 
            expose=expose)
        if expose:
            self.exposed_functions.append(function_name)
            er = ExposedResource(self, function_name)
            function_name_parts = function_name.split("/")
            if len(function_name_parts) > 1:
                if function_name_parts[0] in self.exposed_function_resources:
                    r = self.exposed_function_resources[function_name_parts[0]]
                else:
                    r = Resource()
                    self.exposed_function_resources[function_name_parts[0]] = r
                self.function_resource.putChild(function_name_parts[0], r)
                r.putChild(function_name_parts[1], er)
            else:
                self.function_resource.putChild(function_name_parts[0], er)
            LOGGER.info("Function %s is now available via the HTTP interface." % function_name)
            
    def createReservation(self, function_name, **kwargs):
        if not isinstance(function_name, str):
            for key in self.functions:
                if self.functions[key]["function"] == function_name:
                    function_name = key
                    break
        if function_name not in self.functions:
            raise Exception("Function %s does not exist." % function_name)
        function = self.functions[function_name]
        filtered_kwargs = {}
        for key in function["required_arguments"]:
            if key in kwargs:
                #filtered_kwargs[key] = convertToUTF8(kwargs[key])
                filtered_kwargs[key] = kwargs[key]
            else:
                raise Exception("Required parameter '%s' not found. Required parameters are %s. Optional parameters are %s." % (key, function["required_arguments"], function["optional_arguments"]))
        for key in function["optional_arguments"]:
            if key in kwargs:
                #filtered_kwargs[key] = convertToUTF8(kwargs[key])
                filtered_kwargs[key] = kwargs[key]
        if function["interval"] > 0:
            reserved_arguments = {}
            reserved_arguments["reservation_function_name"] = function_name
            reserved_arguments["reservation_created"] = sdb_now(offset=self.time_offset)
            reserved_arguments["reservation_next_request"] = reserved_arguments["reservation_created"]
            reserved_arguments["reservation_error"] = "0"
            arguments = {}
            arguments.update(reserved_arguments)
            arguments.update(filtered_kwargs)
            uuid = uuid4().hex
            LOGGER.debug("Creating reservation on SimpleDB for %s, %s." % (function_name, uuid))
            a = self.sdb.putAttributes(self.aws_sdb_reservation_domain, uuid, arguments)
            a.addCallback(self._createReservationCallback, function_name, uuid)
            a.addErrback(self._createReservationErrback, function_name, uuid)
            if "call_immediately" in kwargs and not evaluateBoolean(kwargs["call_immediately"]):    
                d = DeferredList([a], consumeErrors=True)
            else:
                LOGGER.debug("Calling %s immediately with arguments:\n%s" % (function_name, PRETTYPRINTER.pformat(filtered_kwargs)))
                self.active_jobs[uuid] = True
                b = self.callExposedFunction(function["function"], filtered_kwargs, function_name, uuid=uuid)
                d = DeferredList([a,b], consumeErrors=True)
            d.addCallback(self._createReservationCallback2, function_name, uuid)
            d.addErrback(self._createReservationErrback2, function_name, uuid)
            return d
        else:
            LOGGER.debug("Calling %s with arguments:\n%s" % (function_name, PRETTYPRINTER.pformat(filtered_kwargs)))
            d = self.callExposedFunction(function["function"], filtered_kwargs, function_name)
            return d
            
    def _createReservationCallback(self, data, function_name, uuid):
        LOGGER.debug("Created reservation on SimpleDB for %s, %s." % (function_name, uuid))
        return uuid

    def _createReservationErrback(self, error, function_name, uuid):
        LOGGER.error("Unable to create reservation on SimpleDB for %s:%s, %s.\n" % (function_name, uuid, error))
        return error

    def _createReservationCallback2(self, data, function_name, uuid):
        for row in data:
            if row[0] == False:
                raise row[1]
        if len(data) == 1:
            return {data[0][1]:{}}
        else:
            return {data[0][1]:data[1][1]}

    def _createReservationErrback2(self, error, function_name, uuid):
        LOGGER.error("Unable to create reservation for %s:%s, %s.\n" % (function_name, uuid, error))
        return error
    
    def showReservation(self, uuid):
        d = self.sdb.getAttributes(self.aws_sdb_reservation_domain, uuid)
        return d
    
    def executeReservation(self, uuid):
        sql = "SELECT * FROM `%s` WHERE itemName() = '%s'" % (self.aws_sdb_reservation_domain, uuid)
        LOGGER.debug("Querying SimpleDB, \"%s\"" % sql)
        d = self.sdb.select(sql)
        d.addCallback(self._executeReservationCallback)
        d.addErrback(self._executeReservationErrback)        
        return d
    
    def _executeReservationCallback(self, data):
        if len(data) == 0:
            raise Exception("Could not find reservation.")
        uuid = data.keys()[0]
        kwargs_raw = {}
        reserved_arguments = {}
        # Load attributes into dicts for use by the system or custom functions.
        for key in data[uuid]:
            if key in self.reserved_arguments:
                reserved_arguments[key] = data[uuid][key][0]
            else:
                kwargs_raw[key] = data[uuid][key][0]
        # Check to make sure the custom function is present.
        function_name = reserved_arguments["reservation_function_name"]
        if function_name not in self.functions:
            raise Exception("Unable to process function %s for UUID: %s" % (function_name, uuid))
            return
        # Check for the presence of all required system attributes.
        if "reservation_function_name" not in reserved_arguments:
            self.deleteReservation(uuid)
            raise Exception("Reservation %s does not have a function name." % uuid)
        if "reservation_created" not in reserved_arguments:
            self.deleteReservation(uuid, function_name=function_name)
            raise Exception("Reservation %s, %s does not have a created time." % (function_name, uuid))
        if "reservation_next_request" not in reserved_arguments:
            self.deleteReservation(uuid, function_name=function_name)
            raise Exception("Reservation %s, %s does not have a next request time." % (function_name, uuid))
        if "reservation_error" not in reserved_arguments:
            self.deleteReservation(uuid, function_name=function_name)
            raise Exception("Reservation %s, %s does not have an error flag." % (function_name, uuid))
        # Load custom function.
        if function_name in self.functions:
            exposed_function = self.functions[function_name]
        else:
            raise Exception("Could not find function %s." % function_name)
            return
        # Check for required / optional arguments.
        kwargs = {}
        for key in kwargs_raw:
            if key in exposed_function["required_arguments"]:
                kwargs[key] = kwargs_raw[key]
            if key in exposed_function["optional_arguments"]:
                kwargs[key] = kwargs_raw[key]
        has_reqiured_arguments = True
        for key in exposed_function["required_arguments"]:
            if key not in kwargs:
                has_reqiured_arguments = False
                raise Exception("%s, %s does not have required argument %s." % (function_name, uuid, key))
        LOGGER.debug("Executing function.\n%s" % function_name)
        return self.callExposedFunction(exposed_function["function"], kwargs, function_name, uuid=uuid)
        
    def _executeReservationErrback(self, error):
        LOGGER.error("Unable to query SimpleDB.\n%s" % error)
    
    
